package httpntlm

import (
	"errors"
	"io"
	"net/http"
	"strings"
	"fmt"

	"github.com/vadimi/go-ntlm/ntlm"
)

var errEmptyNtlm = errors.New("empty NTLM challenge")

// NtlmTransport is implementation of http.RoundTripper interface
type NtlmTransport struct {
	GetCreds func(*http.Request) (string, string, string)
	ModifyRequest func(*http.Request)
	OnResponse func(*http.Request,*http.Response)

	Domain   string
	User     string
	Password string
	http.RoundTripper
	Jar http.CookieJar
}

// RoundTrip method send http request and tries to perform NTLM authentication
func (t NtlmTransport) RoundTrip(req *http.Request) (res *http.Response, err error) {
	if t.GetCreds != nil {
		t.Domain, t.User, t.Password = t.GetCreds(req)
	}

	if t.Domain == "" && t.RoundTripper != nil {
		return t.RoundTripper.RoundTrip(req)
	}

	client := http.Client{}
	if t.RoundTripper != nil {
		client.Transport = t.RoundTripper
	}

	if t.Jar != nil {
		client.Jar = t.Jar
	}

	resp, err := t.ntlmRoundTrip(client, req)
	// retry once in case of an empty ntlm challenge
	if err != nil && errors.Is(err, errEmptyNtlm) {
		return t.ntlmRoundTrip(client, req)
	}

	return resp, err
}

func (t NtlmTransport) ntlmRoundTrip(client http.Client, req *http.Request) (*http.Response, error) {
	fmt.Println("Trying NTLM RoundTrip")

	// first send NTLM Negotiate header 
	r, _ := http.NewRequestWithContext(req.Context(),"GET", req.URL.String(), strings.NewReader(""))
	
	if t.ModifyRequest != nil {
		t.ModifyRequest(r)
	}

	r.Header.Add("Authorization", "NTLM "+encBase64(negotiate()))

	resp, err := client.Do(r)
	if err != nil {
		fmt.Println("NTLM RoundTrip first error",err)
		return nil, err
	}

	fmt.Println("NTLM RoundTrip resp.StatusCode=",resp.StatusCode)

	if err == nil && resp.StatusCode == http.StatusUnauthorized {
		t.OnResponse(r,resp)

		// retrieve Www-Authenticate header from response
		authHeaders := resp.Header.Values("WWW-Authenticate")
		if len(authHeaders) == 0 {
			fmt.Println("NTLM RoundTrip no header")
			return nil, errors.New("WWW-Authenticate header missing")
		}

		// there could be multiple WWW-Authenticate headers, so we need to pick the one that starts with NTLM
		ntlmChallengeFound := false
		var ntlmChallengeString string
		for _, h := range authHeaders {
			if strings.HasPrefix(h, "NTLM") {
				ntlmChallengeFound = true
				ntlmChallengeString = strings.TrimSpace(h[4:])
				break
			}
		}
		if ntlmChallengeString == "" {
			if ntlmChallengeFound {
				fmt.Println("NTLM RoundTrip empty challenge")
				return nil, errEmptyNtlm
			}

			return nil, errors.New("wrong WWW-Authenticate header")
		}

		challengeBytes, err := decBase64(ntlmChallengeString)
		if err != nil {
			fmt.Println("NTLM RoundTrip base64",err)
			return nil, err
		}

		// it's necessary to reuse the same http connection
		// in order to do that it's required to read Body and close it
		_, err = io.Copy(io.Discard, resp.Body)
		if err != nil {
			fmt.Println("NTLM RoundTrip io.Copy error",err)
			return nil, err
		}
		err = resp.Body.Close()
		if err != nil {
			fmt.Println("NTLM RoundTrip response body close error",err)
			return nil, err
		}

		session, err := ntlm.CreateClientSession(ntlm.Version2, ntlm.ConnectionlessMode)
		if err != nil {
			fmt.Println("NTLM RoundTrip create client",err)
			return nil, err
		}

		session.SetUserInfo(t.User, t.Password, t.Domain)

		// parse NTLM challenge
		challenge, err := ntlm.ParseChallengeMessage(challengeBytes)
		if err != nil {
			fmt.Println("NTLM RoundTrip parse challenge",err)
			return nil, err
		}

		err = session.ProcessChallengeMessage(challenge)
		if err != nil {
			fmt.Println("NTLM RoundTrip process challenge",err)
			return nil, err
		}

		// authenticate user
		authenticate, err := session.GenerateAuthenticateMessage()
		if err != nil {
			fmt.Println("NTLM RoundTrip generate auth",err)
			return nil, err
		}

		fmt.Println("NTLM Setting auth header!!")
		// set NTLM Authorization header
		req.Header.Set("Authorization", "NTLM "+encBase64(authenticate.Bytes()))
		return client.Do(req)
	}

	return resp, err
}
